"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.utils import load_json, check_field
from wa_simulator.terrain import WATerrain
from wa_simulator.chrono.system import WAChronoSystem

# Chrono specific imports
import pychrono as chrono
import pychrono.vehicle as veh


def load_chrono_terrain_from_json(filename: str, system: WAChronoSystem):
    """Load a ChTerrain from a json specification file

    Args:
        filename (str): the relative path to a terrain json file
        system (WAChronoSystem): the chrono system used to handle the terrain

    Returns:
        ChTerrain: The loaded terrain
    """
    j = load_json(chrono.GetChronoDataFile(filename))

    # Validate the json file
    check_field(j, 'Terrain', field_type=dict)

    t = j['Terrain']
    check_field(t, 'Input File', field_type=str)
    check_field(t, 'Texture', field_type=str, optional=True)

    terrain = veh.RigidTerrain(system.system, chrono.GetChronoDataFile(t['Input File']))  # noqa

    # Add texture to the terrain, if desired
    if 'Texture' in t:
        texture_filename = chrono.GetChronoDataFile(t['Texture'])

        vis_mat = chrono.ChVisualMaterial()
        vis_mat.SetKdTexture(texture_filename)
        vis_mat.SetSpecularColor(chrono.ChVectorF(0.0, 0.0, 0.0))
        vis_mat.SetFresnelMin(0)
        vis_mat.SetFresnelMax(0.1)

        patch_asset = terrain.GetPatches()[0].GetGroundBody().GetAssets()[0]
        patch_visual_asset = chrono.CastToChVisualization(patch_asset)
        patch_visual_asset.material_list.append(vis_mat)

    return terrain


class WAChronoTerrain(WATerrain):
    """Wrapper for a ChTerrain

    Only support for RigidTerrain is supported for now.

    Args:
        filename (str): the json specification file describing the terrain
        system (WAChronoSystem): system used to create the terrain

    Attributes:
        terrain (RigidTerrain): the created rigid terrain
    """

    def __init__(self, filename, system: WAChronoSystem):
        self.terrain = load_chrono_terrain_from_json(filename, system)

    def synchronize(self, time):
        """Synchronize the terrain at the specified time

        Args:
            time (double): time at which synchronization should occur
        """
        self.terrain.Synchronize(time)

    def advance(self, step):
        """Advance the terrain dynamics by the specified step

        Args:
            step (double): the step at which the simulation should be updated by
        """
        self.terrain.Advance(step)



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

from abc import ABC, abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.system import WASystem
from wa_simulator.core import WAVector
from wa_simulator.utils import load_json, check_field, check_field_allowed_values, get_wa_data_file
from wa_simulator.vehicle import WAVehicle
from wa_simulator.constants import WA_EARTH_RADIUS, WA_PI

# Other Imports
import json
import numpy as np


def load_sensor_suite_from_json(filename: str, manager: 'WASensorManager'):
    """Load a sensor suite from json

    Args:
        filename (str): The json specification file that describes the sensor suite
        manager (WASensorManager): The sensor manager to store all created objects in
    """
    j = load_json(get_wa_data_file(filename))

    # Validate the json file
    check_field(j, 'Type', value='Sensor')
    check_field(j, 'Template', value='Sensor Suite')
    check_field(j, 'Sensors', field_type=list)

    # Load the sensors
    for sensor in j['Sensors']:
        new_sensor = load_sensor_from_json(sensor, manager.vehicle)
        manager.add_sensor(new_sensor)


def load_sensor_from_json(filename: str, vehicle: WAVehicle):
    """Load a sensor from json

    Args:
        filename (str): The json specification file that describes the sensor
        vehicle (WAVehicle): The vehicle each sensor will be attached to
    """
    j = load_json(get_wa_data_file(filename))

    # Validate the json file
    check_field(j, 'Type', value='Sensor')
    check_field(j, 'Template', allowed_values=['IMU', 'GPS'])

    # Check the template and create a sensor based off what it is
    if j['Template'] == 'IMU':
        sensor = WAIMUSensor(vehicle, filename)
    elif j['Template'] == 'GPS':
        sensor = WAGPSSensor(vehicle, filename)
    else:
        raise TypeError(f"{j['Template']} is not an implemented sensor type")

    return sensor


class WASensorManager:
    """Base class that manages sensors.

    Args:
        system (WASystem): The system for the simulation
        filename (str): A json file to load a scene from. Defaults to None (does nothing).

    Attributes:
        system (WASystem): The system for the simulation
        vehicle (WAVehicle): The vehicle each sensor will be attached to
        sensors (list): List of the attached sensors
    """

    # Global filenames for sensor suites
    EGP_SENSOR_SUITE_FILE = "sensors/suites/ev_grand_prix.json"

    def __init__(self, system: WASystem, vehicle: WAVehicle, filename: str = None):
        self.system = system
        self.vehicle = vehicle

        self.sensors = []

        if filename is not None:
            load_sensor_suite_from_json(filename, self)

    def add_sensor(self, sensor: "WASensor"):
        """Add a sensor to the sensor manager"""
        self.sensors.append(sensor)

    def synchronize(self, time):
        """Synchronize the sensor at the specified time

        Args:
            time (float): the time at which the sensors are synchronized to
        """
        for sensor in self.sensors:
            sensor.synchronize(time)

    def advance(self, step):
        """Advance the state of the sensor by the specified time step

        Args:
            step (float): the step to update the sensor by
        """
        for sensor in self.sensors:
            sensor.advance(step)


class WANoiseModel(ABC):
    """Base class for a noise model

    Very similar to ChNoiseModel (https://github.com/projectchrono/chrono/blob/feature/sensor/src/chrono_sensor/ChNoiseModel.h)
    Made because ChSensor currently _requires_ OptiX (NVIDIA proprietary software)
    """

    @abstractmethod
    def add_noise(data: list):
        """Add noise to the data

        Args:
            data (list): The data to add noise to
        """
        pass


class WANoNoiseModel(WANoiseModel):
    """Derived noise model. Does nothing"""

    def add_noise(data: list):
        """Do nothing"""
        pass


class WANormalDriftNoiseModel(WANoiseModel):
    """Derived noise model. Gaussian drifting noise with noncorrelated equal distributions

    Very similar to ChNormalDriftModel (https://github.com/projectchrono/chrono/blob/feature/sensor/src/chrono_sensor/ChNoiseModel.h)
    Made because ChSensor currently _requires_ OptiX (NVIDIA proprietary software)

    Args:
        p (dict): Model properties/parameters taken from a json specification file
    """

    def __init__(self, p: dict):
        # Validate properties
        check_field(p, 'Noise Type', value='Normal Drift')
        check_field(p, 'Update Rate', field_type=float)
        check_field(p, 'Mean', field_type=list)
        check_field(p, 'Standard Deviation', field_type=list)
        check_field(p, 'Bias Drift', field_type=float)
        check_field(p, 'Tau Drift', field_type=float)

        self._load_properties(p)

    def _load_properties(self, p: dict):
        """Private function that loads propeties and sets them to class variables

        Args:
            p (dict): Propeties dictionary
        """
        self.update_rate = p['Update Rate']
        self.mean = WAVector(p['Mean'])
        self.sigma = WAVector(p['Standard Deviation'])
        self.bias_drift = p['Bias Drift']
        self.tau_drift = p['Tau Drift']

        self.bias = WAVector()

    def add_noise(self, data: list):
        """Add noise to the data

        Args:
            data (list): The data to add noise to
        """

        def rand(element):
            return np.random.normal(getattr(self.mean, element, 0), getattr(self.sigma, element, 0))

        eta_a = WAVector([rand('x'), rand('y'), rand('z')])

        eta_b = WAVector()
        if self.tau_drift > 0 and self.bias_drift > 0:
            def rand(): return np.random.normal(0.0, self.drift_bias * np.sqrt(1.0 / (self.update_rate * self.tau_drift)))  # noqa
            eta_b = WAVector([rand(), rand(), rand()])

        self.bias += eta_b
        data += eta_a + self.bias


class WANormalNoiseModel(WANoiseModel):
    """Derived noise model. Gaussian drifting noise with noncorrelated equal distributions

    Very similar to ChNormalModel (https://github.com/projectchrono/chrono/blob/feature/sensor/src/chrono_sensor/ChNoiseModel.h)
    Made because ChSensor currently _requires_ OptiX (NVIDIA proprietary software)

    Args:
        p (dict): Model properties/parameters taken from a json specification file
    """

    def __init__(self, p: dict):
        # Validate properties
        check_field(p, 'Noise Type', value='Normal')
        check_field(p, 'Mean', field_type=list)
        check_field(p, 'Standard Deviation', field_type=list)

        self._load_properties(p)

    def _load_properties(self, p: dict):
        """Private function that loads propeties and sets them to class variables

        Args:
            p (dict): Propeties dictionary
        """
        self.mean = WAVector(p['Mean'])
        self.sigma = WAVector(p['Standard Deviation'])

    def add_noise(self, data: list):
        """Add noise to the data

        Args:
            data (list): The data to add noise to
        """

        def rand(element):
            return np.random.normal(getattr(self.mean, element, 0), getattr(self.sigma, element, 0))

        data += WAVector([rand('x'), rand('y'), rand('z')])


class WASensor(ABC):
    """Base class for a sensor"""

    def __init__(self):
        pass

    @abstractmethod
    def synchronize(self, time):
        """Synchronize the sensor at the specified time

        Args:
            time (float): the time at which the sensors are synchronized to
        """
        pass

    @abstractmethod
    def advance(self, step):
        """Advance the state of the sensor by the specified time step

        Args:
            step (float): the step to update the sensor by
        """
        pass

    @abstractmethod
    def get_data(self):
        """Get the sensor data"""
        pass


class WAIMUSensor(WASensor):
    """Derived sensor class that implements an IMU model

    Args:
        vehicle (WAVehicle): the vehicle that the state data will be extracted from
        filename (str): a json specification file that describes an IMU sensor model
    """

    # Global filenames for imu sensors
    SBG_IMU_SENSOR_FILE = "sensors/models/SBG_IMU.json"

    def __init__(self, vehicle: WAVehicle, filename: str):
        self.vehicle = vehicle

        j = load_json(get_wa_data_file(filename))

        # Validate the json file
        check_field(j, 'Type', value='Sensor')
        check_field(j, 'Template', value='IMU')
        check_field(j, 'Properties', field_type=dict)

        p = j['Properties']
        check_field(p, 'Update Rate', field_type=int)
        check_field(p, 'Noise Model', field_type=dict, optional=True)

        if 'Noise Model' in p:
            check_field(p['Noise Model'], 'Noise Type',
                        allowed_values=["Normal", "Normal Drift"])

        self._load_properties(p)

        self.acc = WAVector()
        self.omega = WAVector()
        self.orientation = WAVector()

    def _load_properties(self, p: dict):
        """Private function that loads properties and sets them to class variables

        Args:
            p (dict): Properties dictionary
        """
        self.update_rate = p['Update Rate']

        self.noise_model = WANoNoiseModel()
        if 'Noise Model' in p:
            n = p['Noise Model']

            if n['Noise Type'] == 'Normal Drift':
                self.noise_model = WANormalDriftNoiseModel(n)
            elif n['Noise Type'] == 'Normal':
                self.noise_model = WANormalNoiseModel(n)
            else:
                raise TypeError(
                    f"{p['Noise Type']} is not an implemented model type")

    def synchronize(self, time):
        """Synchronize the sensor at the specified time

        Args:
            time (float): the time at which the sensors are synchronized to
        """
        self.noise_model.add_noise(self.acc)  # Acceleration (accelerometer)
        self.noise_model.add_noise(self.omega)  # Angular velocity (gyroscope)
        # self.noise_model.add_noise(self.orientation)  # Orientation ("Magnometer") # noqa

    def advance(self, step):
        """Advance the state of the sensor by the specified time step

        Args:
            step (float): the step to update the sensor by
        """
        self.acc = self.vehicle._get_acceleration()
        self.omega = self.vehicle._get_angular_velocity()
        self.orientation = self.vehicle._get_orientation()

    def get_data(self):
        """Get the sensor data

        Returns:
            (WAVector, WAVector, WAQuaternion): Tuple in the form of (acceleration, angular_velocity, orientation)
        """
        return self.acc, self.omega, self.orientation


class WAGPSSensor(WASensor):
    """Derived sensor class that implements an GPS model

    Args:
        vehicle (WAVehicle): the vehicle that the state data will be extracted from
        filename (str): a json specification file that describes an IMU sensor model
    """

    # Global filenames for gps sensors
    SBG_GPS_SENSOR_FILE = "sensors/models/SBG_GPS.json"

    def __init__(self, vehicle: WAVehicle, filename: str):
        self.vehicle = vehicle

        j = load_json(get_wa_data_file(filename))

        # Validate the json file
        check_field(j, 'Type', value='Sensor')
        check_field(j, 'Template', value='GPS')
        check_field(j, 'Properties', field_type=dict)

        p = j['Properties']
        check_field(p, 'Update Rate', field_type=int)
        check_field(p, 'GPS Reference', field_type=list)
        check_field(p, 'Noise Model', field_type=dict, optional=True)

        if 'Noise Model' in p:
            check_field(p['Noise Model'], 'Noise Type',
                        allowed_values=["Normal", "Normal Drift"])

        self._load_properties(p)

    def _load_properties(self, p: dict):
        """Private function that loads properties and sets them to class variables

        Args:
            p (dict): Properties dictionary
        """
        self.update_rate = p['Update Rate']
        self.reference = WAVector(p['GPS Reference'])

        self.noise_model = WANoNoiseModel()
        if 'Noise Model' in p:
            n = p['Noise Model']

            if n['Noise Type'] == 'Normal Drift':
                self.noise_model = WANormalDriftNoiseModel(n)
            elif n['Noise Type'] == 'Normal':
                self.noise_model = WANormalNoiseModel(n)
            else:
                raise TypeError(
                    f"{p['Noise Type']} is not an implemented model type")

        self.pos = WAVector()

    def synchronize(self, time):
        """Synchronize the sensor at the specified time

        Args:
            time (float): the time at which the sensors are synchronized to
        """
        self.noise_model.add_noise(self.pos)

    def advance(self, step):
        """Advance the state of the sensor by the specified time step

        Args:
            step (float): the step to update the sensor by
        """
        self.pos = self.vehicle._get_position()
        self.coord = WAGPSSensor.cartesian_to_gps(self.pos, self.reference)

    def get_data(self):
        """Get the sensor data

        Returns:
            WAVector: The coordinate location of the vehicle in the form of [longitude, latitude, altitude]
        """
        return self.coord

    @staticmethod
    def cartesian_to_gps(coords: WAVector, ref: WAVector):
        """Convert a point from cartesian to gps

        Args:
          coords(WAVector): The coordinate to convert
          ref(WAVector): The "origin" or reference point

        Returns:
          WAVector: The coordinate in the form of[longitude, latitude, altitude]
        """
        lat = (coords.y / WA_EARTH_RADIUS) * 180.0 / WA_PI + ref.y
        lon = (coords.x / (WA_EARTH_RADIUS * np.cos(lat * WA_PI / 180.0))) * 180.0 / WA_PI + ref.x  # noqa
        alt = coords.z + ref.z

        lon = lon + 360.0 if lon < -180.0 else lon - 360.0 if lon > 180.0 else lon

        return WAVector([lon, lat, alt])

    @staticmethod
    def gps_to_cartesian(coords: WAVector, ref: WAVector):
        """Convert a gps coordinate to cartesian given some reference

        Args:
          coords(WAVector): The coordinate to convert
          ref(WAVector): The "origin" or reference point

        Returns:
          WAVector: The x, y, z point in cartesian
        """
        lon = coords.x
        lat = coords.y
        alt = coords.z

        x = ((lon - ref.x) * WA_PI / 180.0) * (WA_EARTH_RADIUS * np.cos(lat * WA_PI / 180.0))  # noqa
        y = ((lat - ref.y) * WA_PI / 180.0) * WA_EARTH_RADIUS
        z = alt - ref.z

        return WAVector([x, y, z])



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

WA_GRAVITY = 9.81  # [m/s^2]
WA_EARTH_RADIUS = 6371000.0  # [m]
WA_PI = 3.141592653589793



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""


from abc import ABC, abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.path import WAPath
from wa_simulator.core import WAVector

# Other imports
import numpy as np
import matplotlib.pyplot as plt


def create_constant_width_track(center: WAPath, width: float):
    """Generates a WAConstantWidthTrack given a centerline and a constant width. Simply "walks" along path and takes the normal from the center at a distance equal to width/2

    Args:
        center (WAPath): The centerline of the track
        width (float): The constant distance between the left and right boundaries (distance between centerline and a boundary is width/2)
    """
    if center.d_points is None:
        raise ValueError(
            'create_constant_width_track: derivative of the centerline has not been initialized')

    left, right = [], []

    points = center.points

    for i in range(len(points)-1):
        ix, iy, iz = points[i]
        d_point = center.d_points[i]
        l = np.linalg.norm(d_point)
        dx = d_point[0] * width / (2 * l)
        dy = d_point[1] * width / (2 * l)
        left.append([ix - dy, iy + dx, iz])
        right.append([ix + dy, iy - dx, iz])

    def close_path_if_necessary(points): return np.array(points) if center.parameters.is_closed and not np.array_equal(
        points[0], points[-1]) else np.vstack((points, points[0]))
    left = close_path_if_necessary(left)
    right = close_path_if_necessary(right)

    # No interpolation to maintain normals
    left_path = WAPath(left, center.parameters)
    right_path = WAPath(right, center.parameters)

    return WATrack(center, left_path, right_path, extras={'width': width})


class WATrack:
    """Base Track object. Basically holds three WAPaths: centerline and two boundaries. This class provides convenience functions so that it is easier to write various track related code

    Args:
        center (WAPath): The centerline of the track.
        left (WAPath): The "left" side boundary. Side is determined by orientation from the centerline by index.
        right (WAPath): The "right" side boundary. Side is determined by orientation from the centerline by index.
        extras (dict): Extra parameters that are used for the track

    Attributes:
        center (WAPath): The centerline of the track.
        left (WAPath): The "left" side boundary. Side is determined by orientation from the centerline by index.
        right (WAPath): The "right" side boundary. Side is determined by orientation from the centerline by index.
    """

    def __init__(self, center: WAPath, left: WAPath, right: WAPath, extras: dict = {}):
        self.center = center
        self.left = left
        self.right = right

        # Will expand out the extras dictionary into class variables
        self.__dict__.update(extras)

    def inside_boundaries(self, point: WAVector):
        """Check whether the passed point is within the track boundaries

        https://stackoverflow.com/a/33155594
        """
        if not isinstance(point, WAVector):
            raise TypeError(
                f'WATrack.inside_boundary: Expects a WAVector, not a {type(point)}.')

        closest_point, idx = self.center.calc_closest_point(point)
        A, B, C = point, WAVector(self.left.points[idx]), WAVector(self.right.points[idx])  # noqa
        a, b, c = (B-C).length, (C-A).length, (A-B).length
        return a**2 + b**2 >= c**2 and a**2 + c**2 >= b**2

    def plot(self):
        """Plot the path. Most likely plotter is matplotlib, but technically anything can be used."""
        self.center.plot('k', show=False)
        self.left.plot('k', show=False)
        self.right.plot('k', show=False)

        # for pl, pr in zip(self.left.points, self.right.points):
        #     plt.plot([pl[0], pr[0]], [pl[1], pr[1]])

        plt.show()



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""


from abc import ABC, abstractmethod  # Abstract Base Class

# WA Simulator
from wa_simulator.core import WAVector

# Other imports
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import splprep, splev
from scipy.spatial.distance import cdist


def load_waypoints_from_csv(filename, **kwargs):
    """Get data points from a csv file. 

    Should be structured as "x,y,z\nx,y,z...". See NumPy.loadtxt for more info on arguments.

    Args:
        filename (str): file to open and read data from

    Returns:
        np.ndarray: a nxm array with each data point in each row
    """
    return np.loadtxt(filename, **kwargs)


def calc_path_length_cummulative(x, y):
    """Get the cummulative distance along a path provided the given x and y position values

    Args:
        x (np.ndarray): x coordinates
        y (np.ndarray): y coordinates

    Returns:
        np.ndarray: the cummulative distance along the path
    """
    return np.cumsum(np.linalg.norm(np.diff(np.column_stack((x, y)), axis=0), axis=1))


def calc_path_curvature(dx, dy, ddx, ddy):
    """Calculate the curvature of a path at each point

    Args:
        dx (np.ndarray): first x derivative
        dy (np.ndarray): first y derivative
        ddx (np.ndarray): second x derivative
        ddy (np.ndarray): second y derivative

    Returns:
        np.ndarray: the curvature at each point
    """
    return (dx * ddy - dy * ddx) / (dx ** 2 + dy ** 2) ** (3 / 2)


class WAPath:
    """Base Path object. To be used to generate paths or trajectories for path planning and/or path following

    Attributes:
        parameters (WAPath.WAPathParameters): The parameters used to interpolate along the path
        waypoints (np.ndarray): The waypoints that the path interpolates about or maintains
        points (np.ndarray): The interpolated points (can be the waypoints)
    """

    class WAPathParameters:
        """Holds the parameters for the interpolated path

        Attributes:
            is_closed (bool): Whether the path is a closed loop. Defaults to True. 
        """
        is_closed = True

    def __init__(self, waypoints=None, parameters=WAPathParameters()):
        # Check points type and shape
        if isinstance(waypoints, list):
            waypoints = np.array(waypoints)
        elif not isinstance(waypoints, np.ndarray):
            raise TypeError(
                'waypoints type is not recognized. List or NumPy array required.')

        self.waypoints = waypoints
        self.points = waypoints
        self.d_points = None

        self.parameters = parameters

    @abstractmethod
    def calc_closest_point(self, pos):
        """Calculate the closest point on the path from the passed position

        Args:
            pos (wa.WAVector): the position to find the closest point on the path to

        Returns:
            wa.WAVector: the closest point on the path
        """
        pass

    @abstractmethod
    def plot(self, *args, show=True, **kwargs):
        """Plot the path

        Args:
            show (bool, optional): show the plot window. Defaults to True.
        """
        plt.plot(self.points[:, 0], self.points[:, 1], *args, **kwargs)
        if show:
            plt.show()


class WASplinePath(WAPath):
    """Spline path implemented with SciPy's splprep and splev methods

    Args:
        waypoints (np.ndarray): the waypoints to fit the spline to
        num_points (int, optional): number of points to interpolate. Defaults to 100.
        smoothness (float, optional): how fit to each point the spline should be. will hit all points by default. Defaults to 0.0.
        is_closed (bool, optional): Is the path a closed loop. Defaults to True.

    Raises:
        TypeError: the waypoints array type is not as expected
    """

    class WASplinePathParameters(WAPath.WAPathParameters):
        """The parameters for a WASplinePath

        Attributes:
            num_points (int): number of points to interpolate along the path. Defaults to 100.
            smoothness (float): How close the path is to the waypoints. Defaults to 0.0 (goes through all the points).
        """
        num_points = 100
        smoothness = 0.0

    def __init__(self, waypoints, parameters=None, **kwargs):
        # Check inputs
        parameters = parameters if parameters is not None else self.WASplinePathParameters()
        allowed_args = {'num_points', 'smoothness', 'is_closed'}
        for key, value in kwargs.items():
            if key not in allowed_args:
                raise ValueError(
                    f'Passed argument {key} is not allowed. Must be any of the following: {allowed_args}')
            setattr(parameters, key, value)

        super().__init__(waypoints, parameters)

        # Interpolate the path
        tck, u = splprep(self.waypoints.T, s=self.parameters.smoothness,
                         per=self.parameters.is_closed)
        u_new = np.linspace(u.min(), u.max(), self.parameters.num_points)

        # Evaluate the interpolation to get values
        self.x, self.y, self.z = splev(u_new, tck, der=0)  # position
        self.dx, self.dy, self.dz = splev(u_new, tck, der=1)  # first derivative # noqa
        self.ddx, self.ddy, self.ddz = splev(u_new, tck, der=2)  # second derivative # noqa

        # store the points for later
        self.points = np.column_stack((self.x, self.y, self.z))
        self.d_points = np.column_stack((self.dx, self.dy, self.dz))

        # Variables for tracking path
        self.last_index = None

    def calc_closest_point(self, pos):
        """Calculate the closest point on the path from the passed position

        Args:
            pos (wa.WAVector): the position to find the closest point on the path to

        Returns:
            wa.WAVector: the closest point on the path
            idx: the index of the point on the path
        """
        dist = cdist(self.points, [pos])
        idx, = np.argmin(dist, axis=0)

        return WAVector([self.x[idx], self.y[idx], self.z[idx]]), idx

    def plot(self, *args, show=True, **kwargs):
        """Plot the path

        Args:
            show (bool, optional): show the plot window. Defaults to True.
        """
        plt.plot(self.x, self.y, *args, **kwargs)
        if show:
            plt.show()

    def calc_length_cummulative(self):
        """Get the cummulative distance along the path

        Returns:
            np.ndarray: Cummulative distance along the path
        """
        return calc_path_length_cummulative(self.x, self.y)

    def calc_curvature(self):
        """Get the curvature at each point on the path

        Returns:
            np.ndarray: Curvature at each point on the path
        """
        return calc_path_curvature(self.dx, self.dy, self.ddx, self.ddy)



"""
Wisconsin Autonomous - https://www.wisconsinautonomous.org

Copyright (c) 2021 wisconsinautonomous.org
All rights reserved.

Use of this source code is governed by a BSD-style license that can be found
in the LICENSE file at the top level of the repo
"""

# WA Simulator
from wa_simulator.core import WAVector, WAQuaternion
from wa_simulator.utils import check_field

# Chrono specific imports
import pychrono as chrono

# ------------------
# vector conversions
# ------------------


def ChVector_to_WAVector(vector: chrono.ChVectorD):
    """Converts a ChVector to a WAVector

    Args:
        vector (ChVector): The vector to convert
    """
    return WAVector([vector.x, vector.y, vector.z])


def ChQuaternion_to_WAQuaternion(quaternion: chrono.ChQuaternionD):
    """Converts a ChQuaternion to a WAQuaternion

    Args:
        quaternion (ChQuaternion): The quaternion to convert
    """
    return WAQuaternion([quaternion.e0, quaternion.e1, quaternion.e2, quaternion.e3])

# ----------------------
# json related utilities
# ----------------------


def ChFrame_from_json(j: dict):
    """Creates a ChFrame from a json object.

    Args:
        j (dict): The json object that will be converted to a ChFrame

    Returns:
        ChFrameD: The frame created from the json object
    """

    # Validate the json file
    check_field(j, 'Position', field_type=list)
    check_field(j, 'Rotation', field_type=list)

    # Do the conversion
    pos = j['Position']
    rot = j['Rotation']
    return chrono.ChFrameD(chrono.ChVectorD(pos[0], pos[1], pos[2]), chrono.ChQuaternionD(rot[0], rot[1], rot[2], rot[3]))


def ChVector_from_list(l: list, vector_type=chrono.ChVectorD):
    """Creates a ChVector from a list

    Args:
        l (list): The list to convert to a ChVector
        vector_type (Type): The ChVector type to convert to
    """
    if len(l) != 3:
        raise TypeError('list must have length 3')

    return vector_type(float(l[0]), float(l[1]), float(l[2]))
